import os.path
import pandas as pd
from numpy.typing import NDArray
from io import StringIO
from bolib3 import np

"""
A bilevel model for toll optimization on a multicommodity transportation network.
Luce Brotcorne, Martine Labbe, Patrice Marcotte, and Gilles Savard. 
Transportation Science, 35(4):345–358, 2001.
"""

# Properties
name: str = "brotcorne2001"
category: str = "transportation"
subcategory: str = ""
datasets: list = [
    os.path.join('bolib3', 'data', 'network_brotcorne_fig1.csv'),
    os.path.join('bolib3', 'data', 'network_highway.csv'),
    os.path.join('bolib3', 'data', 'network_tryangle345.csv'),
]

# Feasible point
x0 = np.array([0.0, 0.0, 0.0, 0.0, 5.0, 0.0, -2.0])
y0 = np.array([0.0, 1.0, 0.0, 1.0, 2.0, 1.0, 1.0])
x = np.array([0.0, 0.0, 0.0, 0.0, 2.0, 2.0])
y = np.array([[0.0, 1.0, 0.0, 0.0, 1.0, 0.0],
              [0.0, 0.0, 1.0, 0.0, 1.0, 1.0]])


# Methods
def F(x, y, data):
    """
    Upper-level objective function
    (non-convex)
    """
    num_commodities: int = data['num_commodities']
    num_arcs: int = data['num_arcs']
    y2d = np.reshape(y, (num_commodities, num_arcs))
    return -np.sum(x*y2d)


def G(x, y, data):
    """
    Upper-level inequality constraints
    (bounds)
    """
    max_toll: NDArray = data['max_toll']
    return np.hstack([
        max_toll - x,
        max_toll + x
    ])


def H(x, y, data):
    """
    Upper-level equality constraints
    (none)
    """
    return np.empty(0)


def f(x, y, data):
    """
    Lower-level objective function
    (linear)
    """
    cost: NDArray = data['cost']
    num_commodities: int = data['num_commodities']
    num_arcs: int = data['num_arcs']
    y2d = np.transpose(np.reshape(y, (num_commodities, num_arcs)))
    return np.sum(np.dot((cost + x), y2d))


def g(x, y, data):
    """
    Lower-level inequality constraints
    (bounds)
    """
    return y


def h(x, y, data):
    """
    Lower-level equality constraints
    """
    num_commodities: int = data['num_commodities']
    num_nodes: int = data['num_nodes']
    num_arcs: int = data['num_arcs']
    inflow: NDArray = data['inflow']
    outflow: NDArray = data['outflow']
    supply_demand: NDArray = data['supply_demand']
    y2d = np.transpose(np.reshape(y, (num_commodities, num_arcs)))
    return (np.matmul(outflow, y2d) - np.matmul(inflow, y2d) - supply_demand).flatten()


def read_data(filepath=datasets[0]):
    """
    Data files should be comma separated in the following format:

    node, commodity_1[, commodity_2][, commodity_3]...
    ***, ***[, ***][, ***]
    ***, ***[, ***][, ***]
    ..., ...[, ...][, ...]

    arc, from, to, cost, max_toll
    ***, ***, ***, ***, ***
    ***, ***, ***, ***, ***
    ..., ..., ..., ..., ...
    """
    with open(filepath, 'r') as f:
        content = f.read()

    # Split at the blank line (two newlines in a row)
    node_part, arc_part = content.strip().split('\n\n')

    # Read each part using pandas and StringIO
    df_nodes = pd.read_csv(StringIO(node_part), skipinitialspace=True)
    df_arcs = pd.read_csv(StringIO(arc_part), skipinitialspace=True)

    # Verify the column headers
    if (not {'node'}.issubset(df_nodes.columns) or
            not {'arc', 'from', 'to', 'cost', 'max_toll'}.issubset(df_arcs.columns)):
        raise ValueError("Tables do not contain the expected columns.")

    # Dimensions of the data
    num_commodities = len(df_nodes.columns) - 1
    num_nodes = len(df_nodes)
    num_arcs = len(df_arcs)

    # Create a mapping from node name to index
    node_to_index = {node_name: idx for idx, node_name in enumerate(df_nodes['node'])}

    # Create inflow and outflow matrices
    inflow = np.zeros((num_nodes, num_arcs))
    outflow = np.zeros((num_nodes, num_arcs))
    for arc_idx, arc in df_arcs.iterrows():
        i = node_to_index[arc['from']]
        j = node_to_index[arc['to']]
        outflow[i, arc_idx] = 1
        inflow[j, arc_idx] = 1

    return {
        'num_commodities': num_commodities,
        'num_nodes': num_nodes,
        'num_arcs': num_arcs,
        'inflow': inflow,
        'outflow': outflow,
        'cost': df_arcs['cost'].to_numpy(),
        'max_toll': df_arcs['max_toll'].to_numpy(),
        'supply_demand': df_nodes.iloc[:, 1:].to_numpy()
    }


def dimensions(key='', data=None):
    """
    If the argument 'key' is not specified, then:
     - a dictionary mapping variable/function names (str) to the corresponding dimension (int) is returned.
    If the first argument 'key' is specified, then:
     - a single integer representing the dimension of the variable/function with the name {key} is returned.
    """
    num_commodities: int = data['num_commodities']
    num_nodes: int = data['num_nodes']
    num_arcs: int = data['num_arcs']
    n = {
        "x": num_arcs,
        "y": num_arcs*num_commodities,
        "F": 1,
        "G": num_arcs*2,
        "H": 0,
        "f": 1,
        "g": num_arcs*num_commodities,
        "h": num_nodes*num_commodities,
    }
    if key in n:
        return n[key]
    return n
